/* Pseudo block device code, copyright (C), 1992, Lawrence Foard 
 * lfoard@turing.org, entropy@world.std.com
 * This code is under the GNU License
 */

/*
 * This is a test version and can lock up under some circumstances,
 * and it still needs support for user space IOCTL calls.
 */

#include <linux/errno.h>
#include <linux/types.h>
#include <linux/fcntl.h>
#include <linux/sched.h>
#include <linux/kernel.h>
#include <asm/segment.h>
#include <asm/system.h>
#include <sys/pseudo.h>
#define MAJOR_NR 11 /* major # for slaves */
#define MASTER_MAJOR 12 /* major # for master */
#define NPSEUDO 16
#include "blk.h"

static int pseudo_size[NPSEUDO];
static int pseudo_slaves[NPSEUDO]; /* number of slaves on this device */
static int pseudo_flags[NPSEUDO];
static int pseudo_serial=0;
static struct request *head_tostart[NPSEUDO],*tocomplete[NPSEUDO];
static struct request *tail_tostart[NPSEUDO];
static struct wait_queue *wait_to_start[NPSEUDO]; 

/* flag defines */
#define P_BUSY   1 /* sleep on pseudo_free */
#define P_MASTERED 2 /* this device has a master */
#define P_CLOSE_REQ 4 /* device requested close */
#define P_CLOSE_ACK 8 /* need to acknowledge it to device still */

#define PRINT

#define GET(minor)\
 {\
  while(pseudo_flags[minor] & P_BUSY)\
   sleep_on(&pseudo_free[minor]);\
  pseudo_flags[minor]|=P_BUSY;\
 }
 
#define RELEASE(minor)\
 {\
  pseudo_flags[minor]&=~P_BUSY;\
  wake_up(&pseudo_free[minor]);\
 }

#ifdef klfgjkl
/* read and write changed to work with kernel memory and file
   table entries. These should probably go in read_write.c */
   
static int k_write(struct file *file,char *buf,unsigned int count)
 {
  unsigned short fs;
  struct inode *inode;
  int ret;
  if (!(inode=file->f_inode))
   return -EBADF;
  if (!(file->f_mode&2))
   return -EBADF;
  if (!count)
   return 0;
  if (file->f_op && file->f_op->write)
   {
    /* save user space segment */
    __asm__("mov %%fs,%0":"=r" (fs));
    /* switch to kernel segment */
    __asm__("mov %0,%%fs"::"r" ((unsigned short) 0x10));
    ret=file->f_op->write(inode,file,buf,count);
    __asm__("mov %0,%%fs"::"r" (fs));
    return ret;
   }
  printk("(K_Write)inode->i_mode=%06o\n\r",inode->i_mode);
  return -EINVAL;
 }

static int k_read(struct file *file,char *buf,unsigned int count)
 {
  struct inode * inode;
  unsigned short fs;
  int ret;
  if (!(inode=file->f_inode))
   return -EBADF;
  if (!(file->f_mode & 1))
   return -EBADF;
  if (!count)
   return 0;
  if (file->f_op && file->f_op->read)
   {
    /* save user space segment */
    __asm__("mov %%fs,%0":"=r" (fs));
    /* switch to kernel segment */
    __asm__("mov %0,%%fs"::"r" ((unsigned short) 0x10));
    ret=file->f_op->read(inode,file,buf,count);
    __asm__("mov %0,%%fs"::"r" (fs));
    return ret;
   }
  printk("(K_Read)inode->i_mode=%06o\n\r",inode->i_mode);
  return -EINVAL;
 }
 
static int k_close(struct file *filp)
 {	
  if (!filp)
   return 0;
  if (filp->f_count==0) 
   {
    printk("Close: file count is 0\n");
    return 0;
   }
  if (filp->f_count>1) 
   {
    filp->f_count--;
    return 0;
   }
  if (filp->f_op && filp->f_op->release)
   filp->f_op->release(filp->f_inode,filp);
  iput(filp->f_inode);
  filp->f_count--;
  return 0;
 }
 
#endif 
 
static void pseudo_end_req(struct request *req,int uptodate);
 
static void test_reset(int minor)
 {
  if (pseudo_slaves[minor] || (pseudo_flags[minor] & P_MASTERED))
   return;
  pseudo_slaves[minor]=0;
  pseudo_flags[minor]=0;
  pseudo_size[minor]=0x7fffffff;
 }
 
static int slave_pseudo_open(struct inode *inode,struct file *filp)
 {
  int minor=MINOR(inode->i_rdev);
  if (minor>=NPSEUDO)
   return -ENODEV;
  if (!(pseudo_flags[minor] & P_MASTERED) || (pseudo_flags[minor] & P_CLOSE_REQ))
   return -EBUSY;
  pseudo_slaves[minor]++;
  return 0;
 }

static void slave_pseudo_release(struct inode *inode, struct file *filp)
 {
  int minor=MINOR(inode->i_rdev);
  pseudo_slaves[minor]--;
  test_reset(minor);
 }

static int slave_pseudo_ioctl(struct inode *inode,struct file *filp,unsigned int cmd,unsigned int param)
 {
/*  int minor=MINOR(inode->i_rdev);
  switch(cmd)
   {
   }*/
  return -EINVAL;
 }

static void pseudo_end_req(struct request *req,int uptodate)
 {
  struct buffer_head * bh;
  req->errors = 0;
  if (!uptodate) 
   {
    printk(DEVICE_NAME " I/O error\n\r");
    printk("dev %04x, sector %d\n\r",req->dev,req->sector);
    req->nr_sectors--;
    req->nr_sectors &= ~1;
    req->sector += 2;
    req->sector &= ~1;		
   }
  if (bh = req->bh) 
   {
    req->bh=bh->b_reqnext;
    bh->b_reqnext=NULL;
    bh->b_uptodate=uptodate;
    unlock_buffer(bh);
    if (bh = req->bh) 
     {
      if (req->nr_sectors < 2) 
       {
        req->nr_sectors = 2;
        printk("end_request: buffer-list destroyed\n");
       }
      req->buffer = bh->b_data;
      /*return;*/
      printk("This weird section of code is actually used!!\n");
     }
   }
  wake_up(&req->waiting);
  req->dev=-1;
  wake_up(&wait_for_request);
 }

static void add_to_complete(int minor,struct request *req)
 {
  req->next=tocomplete[minor];
  tocomplete[minor]=req;
 }

/* we have to do all our own request list handling */
void do_ps_request(void)
 {
  struct request *req;
  int minor;
  while(CURRENT)
   {
    /* move current onto the appropriate tostart list */
    req=CURRENT;
    CURRENT=req->next;
    minor=MINOR(req->dev);
    PRINT("request %d\n",minor);
    /* make sure it is a valid device and has a master */
    if ((minor>=NPSEUDO) || !(pseudo_flags[minor] & P_MASTERED) || (pseudo_flags[minor] & P_CLOSE_REQ))
     {
      PRINT("request invalid\n");
      pseudo_end_req(req,0);
      continue;
     }
    req->next=NULL;
    if (!head_tostart[minor])
     head_tostart[minor]=tail_tostart[minor]=req;
    else
     {
      head_tostart[minor]->next=req;
      head_tostart[minor]=req;
     }
    PRINT("request queued\n");
    wake_up(&wait_to_start[MINOR(req->dev)]);
   }
 }

/* file ops given to additional opens */
static struct file_operations slave_pseudo_fops=
 {
  NULL,
  block_read,
  block_write,
  NULL,
  NULL,
  slave_pseudo_ioctl,
  slave_pseudo_open,
  slave_pseudo_release
 };

/* master functions */

#define MKHEADER(object,cmnd,data_len)\
 {\
  (object).header.size=(data_len)+sizeof(object);\
  (object).header.command=(cmnd);\
  (object).header.serial=pseudo_serial++;\
 }

int master_pseudo_read(struct inode *inode,struct file *filp,char *buf,int count)
 {
  struct request *req;
  int minor=MINOR(inode->i_rdev);
  int offset,size;
  /* wait until there are requests available */
  if (!tail_tostart[minor])
   {
    /* once buffer is empty then we can acknowledge a shutdown request */
    if (pseudo_flags[minor] & P_CLOSE_ACK)
     {
      struct pc_r_close pc_r_close;
      pseudo_flags[minor]&=~P_CLOSE_ACK;
      if (count<sizeof(pc_r_close))
       return -E2BIG;
      MKHEADER(pc_r_close,PC_R_CLOSE,0);
      memcpy_tofs(buf,&pc_r_close,sizeof(pc_r_close));
     }
    if (filp->f_flags & O_NDELAY)
     return(0);
    PRINT("sleeping on wait_to_start\n");
    while(!tail_tostart[minor])
     {
      interruptible_sleep_on(&wait_to_start[minor]);
      if (current->signal & ~current->blocked)
       return -EINTR;
     }
   }
  PRINT("doing read\n");
  /* take top of tostart request list and process */
  req=tail_tostart[minor];
  if (!(tail_tostart[minor]=req->next))
   head_tostart[minor]=NULL;
  offset=req->sector<<9;
  size=req->nr_sectors<<9;
  switch(req->cmd)
   {
    case READ:
     {
      struct pc_read pc_read;
      PRINT("sending read request\n");
      /* make sure the buffer is big enough */
      if (count<sizeof(pc_read))
       {
        pseudo_end_req(req,0);
        return -E2BIG;
       }
      MKHEADER(pc_read,PC_READ,0);
      pc_read.location=offset;
      pc_read.size=size;
      memcpy_tofs(buf,&pc_read,sizeof(pc_read));
      add_to_complete(minor,req);
      return(sizeof(pc_read));
     }
     break;
    case WRITE:
     {
      struct pc_write pc_write;
      PRINT("sending write request\n");
      /* make sure the buffer is big enough */
      if (count<(sizeof(pc_write)+size))
       {
        pseudo_end_req(req,0);
        return -E2BIG;
       }
      MKHEADER(pc_write,PC_WRITE,size);
      pc_write.location=offset;
      pc_write.size=size;
      memcpy_tofs(buf,&pc_write,sizeof(pc_write));
      /* copy buffer data */
      memcpy_tofs(buf+sizeof(pc_write),req->buffer,size);
      add_to_complete(minor,req);
      return(sizeof(pc_write));
     }
     break;
    default: /* disaster */
     printk("corrupt tostart list!\n");
     return -EIO;
   }
 }

/* return a request from the tocomplete list */
static struct request *find_request(int minor,int offset)
 {
  struct request *trail=NULL,*req;
  for(req=tocomplete[minor];req && ((req->sector<<9)!=offset);req=(trail=req)->next);
  if (!req)
   return(NULL);
  if (trail)
   trail->next=req->next;
  else
   tocomplete[minor]=req->next;
  return(req);
 }

/* get response from process */
int master_pseudo_write(struct inode *inode,struct file *filp,char *buf,int count)
 {
  int minor=MINOR(inode->i_rdev);
  struct request *req;
  struct pc_header header;
  if (count<sizeof(header))
   return -EIO;
  memcpy_fromfs(&header,buf,sizeof(header));
  switch(header.command)
   {
    case PC_UNREC: /* something is wrong */
     printk("User level device driver returned PC_UNREQ\n");
     return sizeof(header);
    case PC_CLOSE: /* device requests "nice" shutdown */
     PRINT("PC_CLOSE %d\n",minor);
      /* we need to sync the device without causing a deadlock,
         if we call sync here we stop the device driver (yuck)... */
     pseudo_flags[minor]|=P_CLOSE_REQ | P_CLOSE_ACK;
     return sizeof(header);
    case PC_SIZE: /* set device size */
     {
      struct pc_size pc_size;
      PRINT("PC_SIZE %d\n",minor);
      if (count<sizeof(pc_size))
       return -EIO;
      memcpy_fromfs(&pc_size,buf,sizeof(pc_size));
      pseudo_size[minor]=pc_size.size;
      return sizeof(pc_size);
     }
    case PC_W_REPLY:
     {
      struct pc_w_reply pc_w_reply;
      PRINT("PC_W_REPLY %d\n",minor);
      if (count<sizeof(pc_w_reply))
       return -EIO;
      memcpy_fromfs(&pc_w_reply,buf,sizeof(pc_w_reply));
      if (!(req=find_request(minor,pc_w_reply.write_req.location)))
       {
        printk("PC_W_REPLY for non existant request!!\n");
        return sizeof(pc_w_reply);
       }
      /* see if the write suceeded */
      pseudo_end_req(req,(req->nr_sectors<<9)==pc_w_reply.actually_wrote);
      return sizeof(pc_w_reply);
     }
    case PC_R_REPLY:
     {
      struct pc_r_reply pc_r_reply;
      PRINT("PC_R_REPLY %d\n",minor);
      if (count<sizeof(pc_r_reply))
       return -EIO;
      memcpy_fromfs(&pc_r_reply,buf,sizeof(pc_r_reply));
      if (!(req=find_request(minor,pc_r_reply.read_req.location)))
       {
        printk("PC_R_REPLY for non existant request!!\n");
        return sizeof(pc_r_reply)+pc_r_reply.actually_read;
       }
      /* see if the read suceeded */
      if ((req->nr_sectors<<9)!=pc_r_reply.actually_read)
       {
        pseudo_end_req(req,0);
        return sizeof(pc_r_reply)+pc_r_reply.actually_read;
       }
      /* make sure they gave us enough data */
      if (count<(sizeof(pc_r_reply)+pc_r_reply.actually_read))
       {
        pseudo_end_req(req,0);
        return -EIO;
       }
      memcpy_fromfs(req->buffer,buf+sizeof(pc_r_reply),pc_r_reply.actually_read);
      pseudo_end_req(req,1);
      return sizeof(pc_r_reply)+pc_r_reply.actually_read;
     }
    default: /* invalid command */
     printk("unrecognized command from device driver\n");
     return(header.size);
   }
 }

static int master_pseudo_open(struct inode *inode,struct file *filp)
 {
  int minor=MINOR(inode->i_rdev);
  if (minor>=NPSEUDO)
   return -ENODEV;
  if ((pseudo_flags[minor] & P_MASTERED) || pseudo_slaves[minor])
   return -EBUSY;
  pseudo_flags[minor]|=P_MASTERED;
  return 0;
 }

static void master_pseudo_release(struct inode *inode, struct file *filp)
 {
  struct request *req,*next;
  int minor=MINOR(inode->i_rdev);
  pseudo_flags[minor]&=~P_MASTERED;
  /* return all pending requests with errors */
  for(req=tail_tostart[minor];req;req=next)
   {
    next=req->next;
    pseudo_end_req(req,0);
   }
  for(req=tocomplete[minor];req;req=next)
   {
    next=req->next;
    pseudo_end_req(req,0);
   }
  head_tostart[minor]=tocomplete[minor]=tail_tostart[minor]=NULL;
  test_reset(minor);
 }

static int master_pseudo_ioctl(struct inode *inode,struct file *filp,unsigned int cmd,unsigned int param)
 {
/*  int minor=MINOR(inode->i_rdev);
  switch(cmd)
   {
   }*/
  return -EINVAL;
 }

static struct file_operations master_pseudo_fops=
 {
  NULL,
  master_pseudo_read,
  master_pseudo_write,
  NULL,
  NULL,
  master_pseudo_ioctl,
  master_pseudo_open,
  master_pseudo_release
 };

void pseudo_init(void)
 {
  int a;
  for(a=0;a<NPSEUDO;a++)
   {
    pseudo_slaves[a]=0;
    pseudo_flags[a]=0;
    pseudo_size[a]=0x7fffffff;
    wait_to_start[a]=NULL;
    head_tostart[a]=tocomplete[a]=tail_tostart[a]=NULL;
   }
  blk_size[MAJOR_NR]=pseudo_size;
  blk_dev[MAJOR_NR].request_fn=do_ps_request;
  blkdev_fops[MAJOR_NR]=&slave_pseudo_fops;
  blkdev_fops[MASTER_MAJOR]=&master_pseudo_fops;
 }
