/* Pseudo device library routines Copyright (C), Lawrence Foard, 1992 
 * May be used under the gnu library license
 * I may be contacted at:
 * lfoard@turing.org, entropy@world.std.com, 
 * entropy on the VWIS Linux support BBS 1-508-793-9568 
 * Snail mail:
 * Lawrence Foard
 * Box #108
 * 217 Park Ave.
 * Worcester MA 01609
 */
 
/* TEST VERSION */
#include <stdio.h>
#include <fcntl.h>
#include <sys/pseudo.h>
#include <pseudo_lib.h>

#define BSIZE 5000

#define HEADER ((struct pc_header *)buff)

#define MKHEADER(object,cmnd,data)\
 {\
  (object).header.command=cmnd;\
  (object).header.serial=pseudo->serial++;\
  (object).header.size=sizeof(object)+data;\
 }

PSEUDO *open_pseudo(char *device_name)
 {
  PSEUDO *pseudo=(PSEUDO *) malloc(sizeof(PSEUDO));
  if (strcmp(device_name,"-"))
   {
    if (-1==(pseudo->in_fd=pseudo->out_fd=open(device_name,O_RDWR)))
     {
      perror(device_name);
      return(NULL);
     }
    pseudo->direct=1;
   }
  else
   {
    pseudo->in_fd=0;
    pseudo->out_fd=1;
    pseudo->direct=0;
   }
  pseudo->size=0x7fffffff;
  pseudo->serial=0;
  return(pseudo);
 }

void close_pseudo(PSEUDO *pseudo)
 {
  /* add code here for nice shutdown once the kernel can do it */
  close(pseudo->in_fd);
  close(pseudo->out_fd);
  free(pseudo);
 }

int size_pseudo(PSEUDO *pseudo,int size)
 {
  struct pc_size pc_size;
  MKHEADER(pc_size,PC_SIZE,0);
  pc_size.size=size;
  write(pseudo->out_fd,&pc_size,sizeof(pc_size));
 }

int run_pseudo(PSEUDO *pseudo,int (*read_fun)(),int (*write_fun)(),void *data)
 {
  char buff[BSIZE];
  while(1)
   {
    /* if direct kernel then read as much as possible */
    if (pseudo->direct)
     {
      if (-1==read(pseudo->in_fd,buff,BSIZE))
       {
        perror("read");
        continue;
       }
     }
    /* if reading from a pipe we read the size word first */
    else
     {
#ifdef kflgj
      if (-1==read(dev_fd,buff,sizeof(int)))
       {
        perror("read");
        continue;
       }
      /* now read rest */
      if (HEADER->size>BSIZE)
       {
        fprintf(stderr,"disaster\n");
        exit(1);
       }
#endif       
     }
    switch(HEADER->command)
     {
      case PC_WRITE:
       {
        struct pc_write pc_write=*((struct pc_write *)buff);
        struct pc_w_reply pc_w_reply;
        pc_w_reply.write_req=pc_write;
        MKHEADER(pc_w_reply,PC_W_REPLY,0);
        if (pc_w_reply.actually_wrote=((pc_write.location>=0) && (pc_write.location+pc_write.size)<=pseudo->size) ? pc_write.size : 0)
         {
          pc_w_reply.actually_wrote=write_fun(data,pc_write.location,buff+sizeof(pc_write),pc_write.size);
/*          memcpy(ram+pc_write.location,buff+sizeof(pc_write),pc_write.size);*/
         }
        write(pseudo->out_fd,&pc_w_reply,sizeof(pc_w_reply));
/*        fprintf(stderr,"got write request %d\n",pc_write.location);*/
       }
       break;
      case PC_READ:
       {
        struct pc_read pc_read=*((struct pc_read *)buff);
        struct pc_r_reply pc_r_reply;
        pc_r_reply.read_req=pc_read;
        MKHEADER(pc_r_reply,PC_R_REPLY,0);
        if (pc_r_reply.actually_read=((pc_read.location>=0) && (pc_read.location+pc_read.size)<=pseudo->size) ? pc_read.size : 0)
         {
          MKHEADER(pc_r_reply,PC_R_REPLY,pc_read.size);
          pc_r_reply.actually_read=read_fun(data,pc_read.location,buff+sizeof(pc_r_reply),pc_read.size);
/*          memcpy(buff+sizeof(pc_r_reply),ram+pc_read.location,pc_read.size);*/
          memcpy(buff,&pc_r_reply,sizeof(pc_r_reply));
          write(pseudo->out_fd,buff,sizeof(pc_r_reply)+pc_r_reply.actually_read);
         }
        else
         { 
          MKHEADER(pc_r_reply,PC_R_REPLY,0);
          write(pseudo->out_fd,&pc_r_reply,sizeof(pc_r_reply));
         }
/*        fprintf(stderr,"got read request %d\n",pc_read.location);*/
       }
       break;
      default: 
       {
        struct pc_unrec pc_unrec;
        MKHEADER(pc_unrec,PC_UNREC,0);
        pc_unrec.bad_header=*HEADER;
        write(pseudo->out_fd,&pc_unrec,sizeof(pc_unrec));
        fprintf(stderr,"unrecognized command %d\n",HEADER->command);
       }
     }
   }
 }
 